/* Analyse.cpp */


#include <math.h>

#include "Analyse.h"

#include "TUtils.h"
#include "DupTree.h"
#include "TInject.h"
#include "TreeMan.h"
#include "OpenFile.h"

inline int myround( double x )
  { return ( x<0 ? int(x-.5) : int(x+.5) ); }

_CLASSDEF(GeneWithCost)

//      
class GeneWithCost : public Sortable
  {

  public :

    /* - */
    String name;   //  
    String spec;   //  
    int    tNum;   //  
    String tName;  //  
    double cost;   // ()    
    double leng;   // ()   -  
    double stgr;   //        -"-  
    double cmdT;   //       ( )
    double cost_p; // p-value  'cost'
    double stgr_p; // p-value  'stgr'
    SrtArray ne;   //      

    /* / */
    GeneWithCost( RCString n, RCString s, int num, String tN,
                  double c, double l ) :
      name(n), spec(s), tNum(num), tName(tN), cost(c), leng(l), stgr(0), cmdT(0),
      cost_p(-1), stgr_p(-1), ne() {}
    GeneWithCost( RCGeneWithCost g ) : Sortable(),
      name(g.name), spec(g.spec), tNum(g.tNum), tName(g.tName), cost(g.cost),
      leng(g.leng), stgr(g.stgr), cmdT(g.cmdT), cost_p(g.cost_p), stgr_p(g.stgr_p),
      ne(g.ne) {}
    ~GeneWithCost() {}

    /*    Object'a */
    virtual classType   isA     () const { return UNDEFINED; }
    virtual const char* nameOf  () const { return "GeneWithCost"; }
    virtual bool        isEqual ( RCObject o ) const
      { return ( o.nameOf()==nameOf() && cost==RCGeneWithCost(o).cost ); }
    virtual void        printOn ( ostream& os ) const;

    /*  "" Clonable */
    virtual PClonable clone() const { return new GeneWithCost(ME); }

    /*  "" Sortable */
    virtual bool isLessThan( RCSortable s ) const
      {
        if ( nameOf() != s.nameOf() )
          return false;
        RCGeneWithCost g=(RCGeneWithCost)s;
        if ( stgr > g.stgr ) return true;
        if ( stgr < g.stgr ) return false;
        if ( cost < g.cost ) return true;
        if ( cost > g.cost ) return false;
        if ( name < g.name ) return true;
        if ( name > g.name ) return false;
        return false;
      }

  };

_CLASSDEF(GTreeWithInfo)

//      
class GTreeWithInfo : public Sortable
  {

  public :

    /* - */
    String name;  //  
    int    num;   //  
//    double cost;  //  
    double mCD;   //  
    double mStGR; //   ""
    double mCD_p; // p-value  'mCD'
    double mSG_p; // p-value  'mStGR'

    /* / */
    GTreeWithInfo( RCString n, int num_/*, double c*/ ) :
      name(n), num(num_), /*cost(c),*/ mCD(0), mStGR(0), mCD_p(-1), mSG_p(-1) {}
    GTreeWithInfo( RCGTreeWithInfo gt ) : Sortable(),
      name(gt.name), num(gt.num), /*cost(gt.cost),*/
      mCD(gt.mCD), mStGR(gt.mStGR), mCD_p(gt.mCD_p), mSG_p(gt.mSG_p) {}
    ~GTreeWithInfo() {}

    /*    Object'a */
    virtual classType   isA     () const { return UNDEFINED; }
    virtual const char* nameOf  () const { return "GTreeWithInfo"; }
    virtual bool        isEqual ( RCObject o ) const
      { return ( o.nameOf()==nameOf() && num==RCGTreeWithInfo(o).num ); }
    virtual void        printOn ( ostream& os ) const;

    /*  "" Clonable */
    virtual PClonable clone() const { return new GTreeWithInfo(ME); }

    /*  "" Sortable */
    virtual bool isLessThan( RCSortable s ) const
      {
        if ( nameOf() != s.nameOf() )
          return false;
        RCGTreeWithInfo gt=(RCGTreeWithInfo)s;
        if ( mStGR> gt.mStGR) return true;
        if ( mStGR< gt.mStGR) return false;
        if ( mCD  < gt.mCD  ) return true;
        if ( mCD  > gt.mCD  ) return false;
        if ( name < gt.name ) return true;
        if ( name > gt.name ) return false;
        return false;
      }

  };

_CLASSDEF(Neighbour)

class Neighbour : public Sortable
  {

  public :

    int    dist_g;
    int    dist_s;
    double cost;

    /* / */
    Neighbour( int dg, int ds, double c ) : Sortable(),
      dist_g(dg), dist_s(ds), cost(c) {}
    Neighbour( RCNeighbour n ) : Sortable(),
      dist_g(n.dist_g), dist_s(n.dist_s), cost(n.cost) {}
    ~Neighbour() {}

    operator String() const
      { return ( String(dist_g) + '/' + dist_s ); }

    /*    Object'a */
    virtual classType   isA     () const { return UNDEFINED; }
    virtual const char* nameOf  () const { return "Neighbour"; }
    virtual bool        isEqual ( RCObject o ) const
      { return ( o.nameOf()==nameOf() &&
                 dist_g==RCNeighbour(o).dist_g &&
                 dist_s==RCNeighbour(o).dist_s &&
                 cost==RCNeighbour(o).cost ); }
    virtual void  printOn ( ostream& os ) const
      { os << dist_g << '/' << dist_s; }

    /*  "" Clonable */
    virtual PClonable clone() const { return new Neighbour(ME); }

    /*  "" Sortable */
    virtual bool isLessThan( RCSortable s ) const
      { return ( nameOf()==s.nameOf() && dist_g < RCNeighbour(s).dist_g ); }

  };

_CLASSDEF(Gistogram)

class Gistogram : public Clonable
  {

  public :

    const int N;
    const double MINVAL;
    const double MAXVAL;

  protected :

    int *baskets;
    int nValues;
    int maxNum;

  public :

    /* / */
    Gistogram( int numBaskets, double min, double max ) : Clonable(),
      N(numBaskets), MINVAL(min), MAXVAL(max), nValues(0), maxNum(0)
      { baskets = new int[N]; flush(); }
    Gistogram( RCGistogram g );
    ~Gistogram()
      { delete baskets; }

    void flush();
    int  addValue( double value );

    /*    Object'a */
    virtual classType   isA     () const { return UNDEFINED; }
    virtual const char* nameOf  () const { return "Gistogram"; }
    virtual bool        isEqual ( RCObject ) const
      { return false; /* !!! */ }
    virtual void        printOn ( ostream& os ) const;

    /*  "" Clonable */
    virtual PClonable clone() const { return new Gistogram(ME); }

  };


inline bool isAboveTheThresholds( RCObject o )
  {
    if ( o.nameOf() == "GeneWithCost" )
      {

        IMPORT_REAL( maxCostDelta,  "CostDeltaThreshold" );
        IMPORT_REAL( minDispersion, "DispersionThreshold" );
        IMPORT_REAL( maxCostPValue, "CostDeltaPValueThreshold" );
        IMPORT_REAL( maxDispPValue, "DispersionPValueThreshold" );

        RCGeneWithCost g = (RCGeneWithCost)(o);
        if ( g.cost   <= maxCostDelta  &&
             g.stgr   >= minDispersion &&
             g.cost_p <= maxCostPValue &&
             g.stgr_p <= maxDispPValue )
          return true;
      }
    return false;
  }


/*        */
OK_ERROR analyseInjections( RCClArray gTrees, RCSpeciesTree sTree,
                            RCGenesTable gTable )
  {

    //    
    IMPORT_BOOLEAN ( logAllowed,   "LogAllowed"             );
    IMPORT_REAL    ( costRatio,    "CostRatio"              );
    IMPORT_REAL    ( costRatio2,   "CostRatio2"             );
    IMPORT_INTEGER ( neighRadius,  "NeighbourhoodRadius"    );

    if ( logAllowed )
      tlog << endl << " *   ." << endl;

    //     HTML 
    const String htmlHeader =
      "<html>\n"
      "<head>\n"
      "<meta http-equiv='Content-Type' content='text/html; charset=koi8-r'>\n"
      "<title>     COG'</title>\n"
      "</head>\n"
      "<body>\n";
    const String htmlFooter =
      "\n</body>\n"
      "</html>";

    //   
    ClArray partTitles;
    partTitles.addAt( String(
      " 0.     ."              ), 0 );
    partTitles.addAt( String(
      " 1. &quot;&quot;  ."              ), 1 );
    partTitles.addAt( String(
      " 2.  ,   ."            ), 2 );
    partTitles.addAt( String(
      " 3.  ,   ."          ), 3 );
    partTitles.addAt( String(
      " 4.     ."              ), 4 );
    partTitles.addAt( String(
      " 5.        ." ), 5 );
    partTitles.addAt( String(
      " 6.     ."             ), 6 );
    partTitles.addAt( String(
      " A.    ."         ), 7 );

    //      
    const String tableHeader =
      "  <tr bgcolor='#C0C0C0'>"
      "<th> </th>"
      "<th> </th>"
      "<th> </th>"
      "<th> (%)</th>"
      "<th>p-value </th>"
      "<th></th>"
      "<th>-<br>-<br></th>"
      "<th>p-value -<br>-<br></th>"
      "<th>-<br> .<br> <br></th>"
      "<th> </th></tr>\n";

    //   CostRatio (   )
    globalSettings.set( "CostRatio", costRatio2 );
    if ( logAllowed && costRatio != costRatio2 )
      tlog << " *   Σ  " << costRatio2 << "."
           << endl;

    //        
    TreeInjector injector( gTrees, sTree );      //  TreeIngector'
    CHECK( injector.inject(), "" );              //  
    ClArray   injections=injector.injections();  //   .
    SrtArray  sortedInjections=injections;       //    
    const int N=injections.count();  //     

    //   
    IMPORT_STRING( badGenesFileName, "BadGenesFileName" );
    ofstream bGF;
    if ( !openFile( bGF, badGenesFileName ) )
      return ERROR;

    if ( logAllowed )
      tlog << " *      \"" << badGenesFileName
           << "\"." << endl;

    IMPORT_BOOLEAN( sOut, "SimpleInjectionAnalysisOutput" );

    //  
    bGF << htmlHeader << "\n";

    if ( !sOut )
      {
        bGF << "<a name='begining'></a>\n";
        bGF << "<center><h1>" << badGenesFileName << "</h1></center>\n";
        bGF << "<p> -    TIQMax,    "
            << "        .</p>\n";
    
        //  
        bGF << "<center><h2><a name='contents'></a></h2></center>\n";
        bGF << "<ul>\n";
        for ( int i=partTitles.firstElementIndex(); i<partTitles.count(); i++ )
          bGF << "  <li><a href='#part" << i << "'>" << partTitles[i] << "</a>\n";
        bGF << "</ul>\n";
      }


    /*****************************************************************
     *  0. ޣ              *
     *****************************************************************/

    if ( !sOut )
      {
        bGF << "\n<p><hr><a name='part0'></a>\n";
        bGF << "<center><h2><a href='#contents'>"<< partTitles[0]
            << "</a></h2></center>\n";
    
        //  
        bGF << "<table align='center' bgcolor='#808080' "
            << "cellspacing='1' cellpadding='2'>\n";
        bGF << "  <tr bgcolor='#FFFFFF'>"
            << "<th>No. </th>"
            << "<th>  </th>"
            << "<th><br> </th>"
            << "<th><br> </th>"
            << "</tr>\n";
      }

    SrtArray badGenes;  //      
    SrtArray gTsWInfo;  //      

    // ""     
    for ( int i=0; i<N; i++ ) //   
      { //    i-  

        //          
        RCTreeInjection injection=RCTreeInjection(injections[i]);
        const double initCost=injection.quality();

        SrtArray genes=injection.geneList();  //     
        ClArray  gwc;              //  "  "  
        double   meanCostDelta=0;  //     
        double   meanLength=0;     //     ""

        GTreeWithInfo gTWInfo( injection.name(), i+1 );
        double mStGR=0; //    
        int    gWNbh=0; //       

        //      : ޣ 
        for ( int gn=0; gn<genes.count(); gn++ ) //    
          {

            //    
            RCString geneName=(RCString)(genes[gn]);  //    
            String species=injection.whatSpecies(geneName);  //  
            double length = injection.gene(geneName).distanceFromParent();
              //    

            TreeInjection cutInjection=injection;  //   
            cutInjection.cutGene(geneName);  //    - 
            cutInjection.injectTo(sTree);    //   

            double costDelta=cutInjection.quality()-initCost;  // 
            if ( -1e-12<costDelta && costDelta<1e-12 )  //  -
              costDelta = 0;
            // (  ,   " "???) //

//            if ( costDelta >= 0 ) //     
//              continue;

            //        Σ
            int gi = gwc.add(
              GeneWithCost
                (
                  geneName,                     //  
                  species,                      //  
                  i+1,                          //   
                  injection.name(),             //  
                  costDelta * 100.0 / initCost, //   %-
                  length                        //   
                ) );

            meanCostDelta += (RCGeneWithCost(gwc[gi])).cost;
            meanLength    += (RCGeneWithCost(gwc[gi])).leng;

          } //       

        meanCostDelta /= gwc.count();
        meanLength    /= gwc.count();

        double rmsCostDelta=0;  // .-. -   .   

        //      : "", ""
        for ( int j=0; j<gwc.count(); j++ ) //   ( )  
          {
            RGeneWithCost g=(RGeneWithCost)(gwc[j]);  //    
            rmsCostDelta  += (g.cost-meanCostDelta)*(g.cost-meanCostDelta);
            g.leng/=meanLength;  //      

            //       
            //         

            double   localMeanCostDelta=0;  //    
            double   meanDistOnGT=0;        //  -    
            double   meanDistOnST=0;        //  -    
            SrtArray neighbours;            //    

            //    j-   . 
            RCGeneTree    g_j=injection.gene(g.name);  // j-   
            RCSpeciesTree s_j=sTree.species(g.spec);   //  j-   

            //     
            for ( int k=0; k<gwc.count(); k++ ) //    
              {

                if ( k==j ) continue;  //       

                //    k-   . 
                RCGeneTree    g_k=injection.gene((RGeneWithCost(gwc[k])).name);
                RCSpeciesTree s_k=sTree.species((RGeneWithCost(gwc[k])).spec);

                //       - 
                TreeManipulator tMg_j((RGeneTree)g_j);
                TreeManipulator tMg_k((RGeneTree)g_k);
                int dist=tMg_j.distanceFrom(tMg_k);

                if ( dist>neighRadius ) continue;  //    

                //         
                TreeManipulator tMs_j((RSpeciesTree)s_j);
                TreeManipulator tMs_k((RSpeciesTree)s_k);
                int dist_s=tMs_j.distanceFrom(tMs_k);

                double cost_k=(RGeneWithCost(gwc[k])).cost; //  

                neighbours.add( Neighbour( dist, dist_s, cost_k ) );

                localMeanCostDelta += cost_k;
                meanDistOnGT       += dist;
                meanDistOnST       += dist_s;

              } //      

            g.ne   = neighbours;
            if ( neighbours.count() > 0 )
              {
                localMeanCostDelta /= neighbours.count();
//                meanDistOnGT /= neighbours.count();
//                meanDistOnST /= neighbours.count();
                g.stgr = meanDistOnST / meanDistOnGT;
                mStGR += g.stgr;
                gWNbh ++;
              }

          } //       

        rmsCostDelta = sqrt( rmsCostDelta / (gwc.count()-1) );

        for ( int j=0; j<gwc.count(); j++ )
          {
            RGeneWithCost g=(RGeneWithCost)(gwc[j]);
            g.cmdT = ( g.cost - meanCostDelta ) / rmsCostDelta;
          }

        //  p-values      
        for ( int j=0; j<gwc.count(); j++ ) //   ( )  
          {
            RGeneWithCost g=(RGeneWithCost)(gwc[j]);  //    
            int count_cost=0;
            int count_stgr=0;
            for ( int k=0; k<gwc.count(); k++ ) //    
              {
                RGeneWithCost g_=(RGeneWithCost)(gwc[k]);  //    
                if ( g_.cost <= g.cost )
                  count_cost++;
                if ( g_.stgr >= g.stgr )
                  count_stgr++;
              }
            g.cost_p = double(count_cost) / gwc.count();
            g.stgr_p = double(count_stgr) / gwc.count();
          }

        gTWInfo.mCD   = meanCostDelta;
        gTWInfo.mStGR = ( gWNbh ? mStGR/gWNbh : 0 );

        badGenes.addElements(gwc);  //     
        gTsWInfo.add(gTWInfo);      //     

        if ( !sOut )
          {
            bGF << "  <tr bgcolor='#FFFFFF'>";
            bGF << "<td align='right'>" << (i+1)            << " </td>";
            bGF << "<td>"               << injection.name() << " </td>";
            bGF << "<td>"               << meanCostDelta    << "% </td>";
            bGF << "<td>"               << rmsCostDelta     << "% </td>";
            bGF << "</tr>\n";
          }

//        if ( !((i+1)%10) )
//          cout << (i+1)/10 << "," << flush;

      } //  ""     

    if ( !sOut )
      bGF << "</table>\n";

    //  p-values (    )  
    for ( int i=0; i<N; i++ ) //    
      {
        RGTreeWithInfo gt=(RGTreeWithInfo)(gTsWInfo[i]);  //    
        int count_mCD=0;
        int count_mStGR=0;
        for ( int k=0; k<gTsWInfo.count(); k++ )
          {
            RGTreeWithInfo gt_=(RGTreeWithInfo)(gTsWInfo[k]);
            if ( gt_.mCD <= gt.mCD )
              count_mCD++;
            if ( gt_.mStGR >= gt.mStGR )
              count_mStGR++;
          }
        gt.mCD_p = double( count_mCD   ) / gTsWInfo.count();
        gt.mSG_p = double( count_mStGR ) / gTsWInfo.count();
      }


    /*****************************************************************
     *  1.                                   *
     *****************************************************************/

    if ( !sOut )
      {
        bGF << "\n<p><hr><a name='part1'></a>\n";
        bGF << "<center><h2><a href='#contents'>"<< partTitles[1]
            << "</a></h2></center>\n";

//        bGF << "'' -    ;" << endl
//            << "'' -  - ;" << endl
//            << "'' -  ;" << endl
//            << "'' -      ;" << endl
//            << "'' -    -    ;"
//            << endl
//            << "'' -   ..    -\"-  ;"
//            << endl
//            << "'p-val' - p-value   ( );" << endl
//            << "'' -      (./. )*."
//            << endl << endl;

        //  
        bGF << "<table align='center' bgcolor='#808080' "
            << "cellspacing='1' cellpadding='2'>\n";
        bGF << tableHeader;

        for ( int j=0; j<badGenes.count(); j++ )
          if ( isAboveTheThresholds(badGenes[j]) )
            {
              RCGeneWithCost g=(RCGeneWithCost)(badGenes[j]);
              String name=RCTreeInjection(injections[g.tNum-1]).name();
              bGF << "  <tr bgcolor='#FFFFFF'>" << badGenes[j] << "</tr>\n";
            }
        bGF << "</table>\n";
      }


    /*****************************************************************
     *  2.                                       *
     *****************************************************************/

    if ( sOut )
      {
        bGF << "\n<h2> 1.     </h2>\n";
        bGF << "<table border>\n"
            << "  <tr><th> </th><th> </th><th> </th>"
               "<th>-F<sub>g</sub> </th><th>R<sub>g</sub> </th></tr>\n";
      }
    else
      {
        bGF << "\n<p><hr><a name='part2'></a>\n";
        bGF << "<center><h2><a href='#contents'>"<< partTitles[2]
            << "</a></h2></center>\n";

//        bGF << "\"[].......% (p-val)   (p-val) [] "
//               "\", " << endl
//            << "'' -  -  ;" << endl
//            << "'' -  ;" << endl
//            << "'' -      ;" << endl
//            << "'' -    -    ;"
//            << endl
//            << "'' -   ..    -\"-  ;"
//            << endl
//            << "'p-val' - p-value   ( );" << endl
//            << "'' -      (./. )*;"
//            <<endl
//            << "'' -   ''    ."
//            << endl << endl;

        //  
        bGF << "<table align='center' bgcolor='#808080' "
            << "cellspacing='1' cellpadding='2'>\n";
        bGF << tableHeader;
      }

    for ( int i=0; i<N; i++ ) //   
      {
        RCTreeInjection injection=(RCTreeInjection)(injections[i]);

        if ( !sOut )
          {
            bGF << "  <tr>";
            bGF << "<th colspan='10' bgcolor='#E0E0E0'>"
                << "<a name='treeNo" << (i+1) << "'></a>";
            bGF << (i+1) << "- "
                << ( injection.name()!="" ?
                     String( ": &laquo;" ) + injection.name() + "&raquo;" :
                     String() ) << " </th>";
            bGF << "</tr>\n";
          }

        for ( int j=0; j<badGenes.count(); j++ )
          if ( RCGeneWithCost(badGenes[j]).tNum == (i+1) &&
               isAboveTheThresholds(badGenes[j]) )
          if ( sOut )
            {
              RCGeneWithCost g = RCGeneWithCost(badGenes[j]);
              bGF << "  <tr>";
              bGF << "<td>" << g.tName                    << " </td>"
                  << "<td>" << g.name                     << " </td>"
                  << "<td>" << g.spec                     << " </td>"
                  << "<td>" << String( -g.cost, 5 ) << '%' << " </td>"
                  << "<td>" << String(  g.stgr, 3 )        << " </td>";
              bGF << "</tr>\n";
            }
          else
            bGF << "  <tr bgcolor='#FFFFFF'>" << badGenes[j] << "</tr>\n";
      }

    bGF << "</table>\n";


    /*****************************************************************
     *  3.                                    *
     *****************************************************************/

    if ( !sOut )
      {
        bGF << "\n<p><hr><a name='part3'></a>\n";
        bGF << "<center><h2><a href='#contents'>"<< partTitles[3]
            << "</a></h2></center>\n";

        //  
        bGF << "<table align='center' bgcolor='#808080' "
            << "cellspacing='1' cellpadding='2'>\n";
        bGF << tableHeader;

        for ( int n=0; n<gTable.numberOfSpecies(); n++ )
          {
            const String curSpec=gTable.species(n);
            bGF << "  <tr><th colspan='10' bgcolor='#E0E0E0'>"
                << "<a name=spec_" << curSpec << "></a>"
                << " &laquo;" << curSpec << "&raquo;"
                << " </th></tr>\n";
            for ( int j=0; j<badGenes.count(); j++ )
              if ( RCGeneWithCost(badGenes[j]).spec == curSpec &&
                   isAboveTheThresholds(badGenes[j]) )
                bGF << "  <tr bgcolor='#FFFFFF'>" << badGenes[j] << "</tr>\n";
          }

        bGF << "</table>\n";
      }


    if ( !sOut )
    { //  ޣ     

      /*****************************************************************
       *  4.                                   *
       *****************************************************************/

      bGF << "\n<p><hr><a name='part4'></a>\n";
      bGF << "<center><h2><a href='#contents'>"<< partTitles[4]
          << "</a></h2></center>\n";

      double *c=new double[N];   //   
      double *cn=new double[N];  //    
      double meanInjectionCost=0;
      double meanInjectionNormCost=0;

      for ( int n=0; n<N; n++ )  //  ,  .
        {
          RCTreeInjection injection_n=RCTreeInjection(injections[n]);
          c[n]  = injection_n.quality();
          cn[n] = injection_n.quality() / injection_n.numberOfLeaves();
          meanInjectionCost     += c[n]/N;
          meanInjectionNormCost += cn[n]/N;
        }

      double costDispersion=0;
      double normCostDispersion=0;
      for ( int n=0; n<N; n++ )  //  
        costDispersion +=
          ( c[n]  - meanInjectionCost     )*( c[n]  - meanInjectionCost     ),
        normCostDispersion +=
          ( cn[n] - meanInjectionNormCost )*( cn[n] - meanInjectionNormCost );
      costDispersion     = sqrt(costDispersion/(N-1/*?*/));
      normCostDispersion = sqrt(normCostDispersion/(N-1/*?*/));

      bGF << "<table align='center'>\n";
      bGF << "  <tr><td>  : </td><td>"
          << meanInjectionCost     << " </td></tr>\n";
      bGF << "  <tr><td>ݣ   : </td><td>"
          << costDispersion        << " </td></tr>\n";
      bGF << "  <tr><td>   : </td><td>"
          << meanInjectionNormCost << " </td></tr>\n";
      bGF << "  <tr><td>ݣ    : </td><td>"
          << normCostDispersion    << " </td></tr>\n";
      bGF << "</table>\n<p>\n";

      //  
      bGF << "<table align='center' bgcolor='#808080' "
          << "cellspacing='1' cellpadding='2'>\n";
      bGF << "<tr bgcolor='#FFFFFF'>"
          << "<th>No<br> </th>"
             "<th>  </th>"
             "<th>No<br><br> </th>"
             "<th> </th>"
             "<th><br> </th>"
             "<th>.<br>  </th>"
             "<th>. <br>  . </th>"
             "<th> / &sigma; </th>"
             "<th><br> </th>"
          << "</tr>\n";

//      bGF << "      :" << endl
//          << "\"#. 'name' (c/N)=nc, v, p -- mX\", " << endl
//          << "#    -  ;" << endl
//          << "name -  ;" << endl
//          << "c    -  ;" << endl
//          << "N    -    ;" << endl
//          << "nc   - . ;" << endl
//          << "v    -    ;" << endl
//          << "p    - v/sigma" << endl
//          << "mX   -    \"\"." << endl << endl;

      for ( int n=0; n<N; n++ )
        {
          RCTreeInjection injection=RCTreeInjection(injections[n]);

          double meanStgr=0;
          int    genesNum=0;
          for ( int j=0; j<badGenes.count(); j++ )
            if ( RCGeneWithCost(badGenes[j]).tNum-1 == n &&
                 RCGeneWithCost(badGenes[j]).ne.count()>0 )
              meanStgr += RCGeneWithCost(badGenes[j]).stgr, genesNum++;
          if ( genesNum > 0 )
            meanStgr/=genesNum;

          int place=sortedInjections.firstIndexOf(injection);

          bGF << "  <tr bgcolor='#FFFFFF'>";
          bGF << "<td nowrap align='right'>" << (n+1)              <<". </td>"
              << "<th nowrap><a href='#treeNo" << (n+1) << "'>"
              << injection.name()                              << "</a> </th>"
              << "<td nowrap align='right'>" << (place+1)           <<" </td>"
              << "<td nowrap>"               << injection.quality() <<" </td>"
              << "<td nowarp align='right'>" << injection.numberOfLeaves()
                                                                    <<" </td>"
              << "<td nowarp>" << cn[n]                             <<" </td>"
              << "<td nowrap>" << (cn[n]-meanInjectionNormCost)     <<" </td>"
              << "<td nowrap>" << (cn[n]-meanInjectionNormCost) /
                                  normCostDispersion                <<" </td>"
              << "<td nowrap>" << meanStgr                          <<" </td>";
          bGF << "</tr>\n";
        }
      
      bGF << "</table>\n";

    } //   ޣ     


    { //         

      /*****************************************************************
       *  5.                *
       *****************************************************************/

      if ( sOut )
        bGF << "\n<h2> 2.     &lt;R<sub>g</sub>&gt;</h2>\n";
      else
        {
          bGF << "\n<p><hr><a name='part5'></a>\n";
          bGF << "<center><h2><a href='#contents'>"<< partTitles[5]
              << "</a></h2></center>\n";
        }

      //    ...
      double mean_mCD=0;
      double mean_mStGR=0;
      for ( int i=0; i<gTsWInfo.count(); i++ )
        {
          RCGTreeWithInfo gt=RCGTreeWithInfo(gTsWInfo[i]);
          mean_mCD   += gt.mCD;
          mean_mStGR += gt.mStGR;
        }
      mean_mCD   /= gTsWInfo.count();
      mean_mStGR /= gTsWInfo.count();

      //  ݣ   
      double rms_mCD=0;
      double rms_mStGR=0;
      for ( int i=0; i<gTsWInfo.count(); i++ )
        {
          RCGTreeWithInfo gt=RCGTreeWithInfo(gTsWInfo[i]);
          rms_mCD   += ( gt.mCD   - mean_mCD   )*( gt.mCD   - mean_mCD   );
          rms_mStGR += ( gt.mStGR - mean_mStGR )*( gt.mStGR - mean_mStGR );
        }
      rms_mCD   = sqrt( rms_mCD   / (gTsWInfo.count()-1) );
      rms_mStGR = sqrt( rms_mStGR / (gTsWInfo.count()-1) );

      if ( !sOut )
        {
          bGF << "<table align='center'>\n";
          bGF << "  <tr><td> ( )   : </td><td>"
              << mean_mStGR << " </td></tr>\n";
          bGF << "  <tr><td>ݣ   : </td><td>"
              << rms_mStGR << " </td></tr>\n";
          bGF << "  <tr><td> (   )   : </td><td>"
              << mean_mCD << " </td></tr>\n";
          bGF << "  <tr><td>ݣ    : </td><td>"
              << rms_mCD << " </td></tr>\n";
          bGF << "</table>\n<p>\n";
        }

      //  
      if ( sOut )
        bGF << "<table border>\n"
            << "  <tr>"
            << "<th>No.  </th>"
            << "<th>  </th>"
            << "<th>-&lt;F<sub>g</sub>&gt; </th>"
            << "<th>&lt;R<sub>g</sub>&gt; </th>"
            << "</tr>\n";
      else
          bGF << "<table align='center' bgcolor='#808080' "
              << "cellspacing='1' cellpadding='2'>\n"
              << "  <tr bgcolor='#C0C0C0'>"
              << "<th>No.  </th>"
              << "<th>  </th>"
              << "<th><br>-<br> </th>"
              << "<th>p-value<br><br>-<br> </th>"
              << "<th><br> </th>"
              << "<th>p-value<br><br> </th>"
              << "</tr>\n";

      for ( int i=0; i<gTsWInfo.count(); i++ )
        if ( sOut )
          {
            RCGTreeWithInfo gt = RCGTreeWithInfo(gTsWInfo[i]);
            bGF << "  <tr>";
            bGF << "<td>" << gt.num                << " </td>"
                << "<td>" << gt.name               << " </td>"
                << "<td>" << String( -gt.mCD,   5 ) << " </td>"
                << "<td>" << String(  gt.mStGR, 3 ) << " </td>";
            bGF << "</tr>\n";
          }
        else
          bGF << "  <tr bgcolor='#FFFFFF'>" << gTsWInfo[i] << "</tr>\n";

      bGF << "</table>\n";

    } //          


    if ( !sOut )
    { //   
    
    /*****************************************************************
     *  6.                                          *
     *****************************************************************/

    bGF << "\n<p><hr><a name='part6'></a>\n";
    bGF << "<center><h2><a href='#contents'>"<< partTitles[6]
        << "</a></h2></center>\n";

    //      
    double minCost=0;
    double maxCost=0;
    for ( int j=0; j<badGenes.count(); j++ )
      if ( RCGeneWithCost(badGenes[j]).cost <= 0 )
        {
          double cost=RCGeneWithCost(badGenes[j]).cost;
          if ( cost<minCost || !j ) minCost=cost;
          if ( cost>maxCost || !j ) maxCost=cost;
        }

    for ( int i=0; i<N; i++ ) //   
      {
        RCTreeInjection injection=(RCTreeInjection)(injections[i]);

        bGF << "<center><table width='90%' bgcolor='#E0E0E0'><tr><th>\n";
        bGF << (i+1) << "- "
            << ( injection.name()!="" ?
                 String( ": &laquo;" ) + injection.name() + "&raquo;" :
                 String() );
        bGF << "</th></tr>\n";

        Gistogram gistogram( 16, minCost, maxCost );
        for ( int j=0; j<badGenes.count(); j++ )
          if ( RCGeneWithCost(badGenes[j]).tNum == (i+1) &&
               RCGeneWithCost(badGenes[j]).cost <= 0 )
            gistogram.addValue( RCGeneWithCost(badGenes[j]).cost );

        bGF << "<tr><td>\n";
        bGF << gistogram << "\n";
        bGF << "</td></tr></table></center>\n";
        bGF << "<br>\n";

      }

    } //    


    /*****************************************************************
     *  A.                   *
     *****************************************************************/

    if ( !sOut )
      {
        bGF << "\n<p><hr><a name='part7'></a>\n";
        bGF << "<center><h2><a href='#contents'>"<< partTitles[6]
            << "</a></h2></center>\n";

        SrtArray abbreviations;
        abbreviations.add( String( " -  " ) );
        abbreviations.add( String( " -  " ) );
        abbreviations.add( String( " -   " ) );
        abbreviations.add( String( " -   " ) );
        abbreviations.add( String( " -  " ) );
        abbreviations.add( String( " -  " ) );
        abbreviations.add( String( " -   ()" ) );
        abbreviations.add( String( " -   ()" ) );
        abbreviations.add( String( " -  " ) );

        bGF << "<ul>\n";
        for ( int i=0; i<abbreviations.count(); i++ )
          bGF << "  <li>" << abbreviations[i] << "</li>\n";
        bGF << "</ul>\n";
      }


    bGF << htmlFooter;

    bGF.close();



    /***       ... ******************/
    //        ..
    //    ..

    if ( !GLOBAL_BOOLEAN("AnalyseEvolutionaryEvents") )
      return OK;

    //     ...

    IMPORT_STRING( resultFileName, "ResultFileName" );
    String resultFileExt = "";
    String resultFileNam = resultFileName;
    if ( resultFileName.contains('.') and
         resultFileName.lastIndexOf('.') >
         resultFileName.lastIndexOf('/') )
      {
        resultFileExt = resultFileName.substring(    resultFileName.lastIndexOf('.') );
        resultFileNam = resultFileName.substring( 0, resultFileName.lastIndexOf('.') );
      }
    String treeFileNGDup    = resultFileNam + "-1a-non-gain-dup"     + resultFileExt;
    String treeFileNGSumDup = resultFileNam + "-1b-non-gain-sum-dup" + resultFileExt;
    String treeFileNGLosses = resultFileNam + "-1c-non-gain-losses"  + resultFileExt;
    String treeFileGDup     = resultFileNam + "-2a-gain-dup"         + resultFileExt;
    String treeFileGSumDup  = resultFileNam + "-2b-gain-sum-dup"     + resultFileExt;
    String treeFileGLosses  = resultFileNam + "-2c-gain-losses"      + resultFileExt;
    String treeFileGains    = resultFileNam + "-3-gains"             + resultFileExt;

    if ( logAllowed )
      tlog << " *       ." << endl;

    ofstream resultFile;

    /*  " " */
    DuplicationsTree dTree(sTree);

    /*   */
    TreeInjector injector2(gTrees,dTree);
    injector2.inject();

    tlog << endl << " * Non-Gain :" << endl;
    dTree.defineNumberOfDuplications( injector2 );  //  
    tlog << " 1)  :" << endl;
    tlog << dTree << ';' << endl;
    if ( !openFile( resultFile, treeFileNGDup ) )
      return ERROR;
    resultFile << dTree << ';' << endl;
    resultFile.close();
    dTree.defineSumOfDuplications( injector2 );
    tlog << " 2)   :" << endl;
    tlog << dTree << ';' << endl;
    if ( !openFile( resultFile, treeFileNGSumDup ) )
      return ERROR;
    resultFile << dTree << ';' << endl;
    resultFile.close();
    dTree.defineSumOfLosses( injector2 );
    const uns losses_NG = dTree.numberOfDuplications();
    tlog << " 3)   :" << endl;
    tlog << dTree << ';' << endl;
    if ( !openFile( resultFile, treeFileNGLosses ) )
      return ERROR;
    resultFile << dTree << ';' << endl;
    resultFile.close();

    uns numAbnormalGenes = 0;
    tlog << endl << " *   ." << endl;
    for ( int j=0; j<badGenes.count(); j++ ) //   
      if ( isAboveTheThresholds(badGenes[j]) ) //  
        { //     
          RCGeneWithCost g=RCGeneWithCost(badGenes[j]);
          if ( removeGene( RTreeInjection((injector2.injections())[g.tNum-1]),
                           g.name ) == ERROR )
            {
              cerr << "    '" << g.name << "'   ." << endl;
              return ERROR;
            }
          numAbnormalGenes++;
        }

    injector2.inject();

    if ( !numAbnormalGenes )
      {
        tlog << endl << " *        ." << endl;
        return OK;
      }
    tlog << " *      (" << numAbnormalGenes << ")."
         << endl;

    tlog << endl << " * Gain :" << endl;
    dTree.defineNumberOfDuplications( injector2 );  //  
    tlog << " 1)  :" << endl;
    tlog << dTree << ';' << endl;
    if ( !openFile( resultFile, treeFileGDup ) )
      return ERROR;
    resultFile << dTree << ';' << endl;
    resultFile.close();
    dTree.defineSumOfDuplications( injector2 );
    tlog << " 2)   :" << endl;
    tlog << dTree << ';' << endl;
    if ( !openFile( resultFile, treeFileGSumDup ) )
      return ERROR;
    resultFile << dTree << ';' << endl;
    resultFile.close();
    dTree.defineSumOfLosses( injector2 );
    const uns losses_G = dTree.numberOfDuplications();
    tlog << " 3)   :" << endl;
    tlog << dTree << ';' << endl;
    if ( !openFile( resultFile, treeFileGLosses ) )
      return ERROR;
    resultFile << dTree << ';' << endl;
    resultFile.close();

    //    
    dTree.clearNumberOfDuplications();
    for ( int j=0; j<badGenes.count(); j++ ) //   
      if ( isAboveTheThresholds(badGenes[j]) ) //  
        {
          RCGeneWithCost g=RCGeneWithCost(badGenes[j]);
          TreeManipulator treeMan(dTree.species(g.spec));
          do
            RDuplicationsTree(treeMan.host()).numberOfDuplications(
              RDuplicationsTree(treeMan.host()).numberOfDuplications() + 1 );
          while ( treeMan.goToParent() != ERROR );
        }
    tlog << " *  :" << endl;
    tlog << dTree << ';' << endl;
    if ( !openFile( resultFile, treeFileGains ) )
      return ERROR;
    resultFile << dTree << ';' << endl;
    resultFile.close();

    tlog << " !    1  : " <<
      String( double(losses_NG-losses_G)/numAbnormalGenes, 2 ) << endl;
    cout << "(" << String( double(losses_NG-losses_G)/numAbnormalGenes, 2 ) << ") ";

    globalSettings.set( "CostRatio", costRatio );

    return OK;

  }


void GeneWithCost::printOn( ostream& os ) const
  {
    os << "<td nowrap>"  << name                              << " </td>"
       << "<td nowrap align='center'><a href='#spec_"
       << spec << "'>"  << spec  << "</a>"                    << " </td>"
       << "<td nowrap><a href='#treeNo" << tNum << "'>"
       << tName << "</a>"                                     << " </td>"
       << "<td nowrap>"  << String( cost, 5 ) << '%'          << " </td>"
       << "<td nowrap align='center'>" << String( cost_p, 2 ) << " </td>"
       << "<td nowrap>"  << String( leng, 5 )                 << " </td>"
       << "<td nowrap align='center'>" << String(stgr,3)      << " </td>"
       << "<td nowrap align='center'>" << String(stgr_p,2)    << " </td>"
       << "<td nowrap>"  << String(cmdT,3) << "&sigma;"       << " </td>"
       << "<td><font size='2'>";
    for ( int i=0; i<ne.count(); i++ )
      os << String(RCNeighbour(ne[i])) << ( i<ne.count()-1 ? ", " : "" );
    os << "</font> </td>";
  }


void GTreeWithInfo::printOn( ostream& os ) const
  {
    os << "<td nowrap>"  << num                              << " </td>"
       << "<td nowrap><a href='#treeNo" << num << "'>"
       << name << "</a>"                                     << " </td>"
       << "<td nowrap align='center'>" << String( mStGR, 3 ) << " </td>"
       << "<td nowrap align='center'>" << String( mSG_p, 2 ) << " </td>"
       << "<td nowrap align='center'>" << String( mCD,   5 ) << " </td>"
       << "<td nowrap align='center'>" << String( mCD_p, 2 ) << " </td>";
  }


Gistogram::Gistogram( RCGistogram g ) : Clonable(),
  N(g.N), MINVAL(g.MINVAL), MAXVAL(g.MAXVAL), maxNum(g.maxNum)
  {
    for ( int i=0; i<N; i++ )
      baskets[i] = g.baskets[i];
  }

void Gistogram::flush()
  {
    for ( int i=0; i<N; i++ )
      baskets[i]=0;
    maxNum = nValues = 0;
  }

#include <math.h>

int Gistogram::addValue( double value )
  {
    if ( value < MINVAL || value > MAXVAL )
      return ERROR;

    int n=int(floor( (value-MINVAL) / (MAXVAL-MINVAL+1e-12) * N ));
    baskets[n]++;
    nValues++;
    if ( baskets[n] > maxNum )
      maxNum = baskets[n];

    return n;
  }

void Gistogram::printOn( ostream& os ) const
  {
    const double factor = 100.0/maxNum;

    os << "<table width='100%' cellpadding='0' cellspacing='1' border='0' bgcolor='#D0E0F0'>\n";

    os << "  <tr><th><table width='100%' cellpadding='0' cellspacing='1' border='0' bgcolor='black'><tr>\n";
    for ( int i=0; i<10; i++ )
      os << "    <td width='10%' bgcolor='#B0C0D0' align='left'><font size='2'>" << i*10 << "</font></td>\n";
    os << "  </tr></table></th></tr>\n";

    os << "  <tr><td>\n";
    for ( int i=0; i<N; i++ )
      {
        int    height=myround(baskets[i]*factor);
        double value = MINVAL + (MAXVAL-MINVAL)/N*(i+.5);
//        double percent=double(baskets[i])/nValues*100;
        String label;
        if ( baskets[i] )
          label = String(baskets[i]) + " X " + String(value,3);
        String nbsp  = "&nbsp;";
        os << "    <table width='100%' cellpadding='0' cellspacing='1' border='0' bgcolor='#D0E0F0'><tr>\n";
        os << "      <td width='" << height << "%' align='right' bgcolor='#4080C0'><font size='1'><b>" << (height>=50?label:nbsp) << "</b></font></td>\n";
        os << "      <td width='" << (100-height) << "%'><font size='1'><b>" << (height<50?label:nbsp) << "</b></font></td>\n";
        os << "    </tr></table>\n";
      }
    os << "  </td></tr>\n";

    os << "  <tr><th><table width='100%' cellpadding='0' cellspacing='1' border='0' bgcolor='black'><tr>\n";
    for ( int i=0; i<10; i++ )
      os << "    <td width='10%' bgcolor='#B0C0D0' align='left'><font size='2'>" << i*10 << "</font></td>\n";
    os << "  </tr></table></th></tr>\n";

    os << "  <tr><td>\n";
    os << "    N=" << nValues << ", min=" << String(MINVAL,4) <<
                                 ", max=" << String(MAXVAL,4) << ".\n";
    os << "  </td></tr>\n";

    os << "</table>\n";

  }
