/* Settings.h */

#ifndef __SETTINGS_H
#define __SETTINGS_H

#include "ClString.h"
#include "SrtArray.h"

_CLASSDEF(Setting)
_CLASSDEF(Settings)
_CLASSDEF(RealSetting)
_CLASSDEF(StringSetting)
_CLASSDEF(BooleanSetting)
_CLASSDEF(IntegerSetting)

abstract class Setting : public Sortable
  {

  public :

    /* / */
    Setting( RCString aKey ) : _key(aKey) {}
    virtual ~Setting() {}

    virtual bool isRealSetting    () const { return false; }
    virtual bool isStringSetting  () const { return false; }
    virtual bool isBooleanSetting () const { return false; }
    virtual bool isIntegerSetting () const { return false; }

    virtual OK_ERROR setValue( RCString newValue ) = 0;
    virtual OK_ERROR setValue( double   newValue ) = 0;
    virtual OK_ERROR setValue( bool     newValue ) = 0;
    virtual OK_ERROR setValue( int      newValue ) = 0;
    /*   */ OK_ERROR setValue( const char* newValue )
      { setValue( String(newValue) ); return OK; }

    /*   */ String key          () const { return _key; }
    virtual String stringValue  () const = 0;
    double  realValue    () const;
    bool    booleanValue () const;
    int     integerValue () const;

    virtual void restoreDefault() = 0;

    RCRealSetting    toRealSetting    () const;
    RCStringSetting  toStringSetting  () const;
    RCBooleanSetting toBooleanSetting () const;
    RCIntegerSetting toIntegerSetting () const;
    RRealSetting     toRealSetting    ()
      { return RRealSetting( RCSetting(ME).toRealSetting() );       }
    RStringSetting   toStringSetting  ()
      { return RStringSetting( RCSetting(ME).toStringSetting() );   }
    RBooleanSetting  toBooleanSetting ()
      { return RBooleanSetting( RCSetting(ME).toBooleanSetting() ); }
    RIntegerSetting  toIntegerSetting ()
      { return RIntegerSetting( RCSetting(ME).toIntegerSetting() ); }

    /*    Object'a */
    virtual void      printOn ( ostream& os ) const
      { os << key() << "=" << stringValue(); }

    /*  "" Sortable */
    virtual bool isLessThan( RCSortable s ) const
      { return ( isA()<s.isA() || isA()==s.isA() && _key.isLessThan(RCSetting(s)._key) ); }

  private :

    const String _key;

    friend class Settings;

  };

class RealSetting : public Setting
  {

  public :

    /* / */
    RealSetting( RCString aKey, double defValue ) : Setting( aKey ),
      _defaultValue( defValue ),
      _value( defValue )
      {}
    virtual ~RealSetting() {}

    virtual bool isRealSetting () const { return true; }

    virtual OK_ERROR setValue( double   newValue )
      { _value = newValue; return OK; }
    virtual OK_ERROR setValue( RCString newValue )
      { _value = double(newValue); return OK; }
    virtual OK_ERROR setValue( bool     newValue )
      { setValue( double(newValue) ); return ERROR; }
    virtual OK_ERROR setValue( int      newValue )
      { return setValue( double(newValue) ); }

    double value        () const { return _value;        }
    double defaultValue () const { return _defaultValue; }

    virtual String stringValue    () const { return String(_value); }
    virtual void   restoreDefault () { setValue(_defaultValue);     }

    /*    Object'a */
    virtual classType   isA     () const { return realSettingClass; }
    virtual const char* nameOf  () const { return "RealSetting"; }
    virtual bool        isEqual ( RCObject object ) const;

    /*  "" Clonable */
    virtual PClonable clone() const { return new RealSetting(ME); }

  private :

    const double _defaultValue;
    /* */ double _value;

  };

class StringSetting : public Setting
  {

  public :

    /* / */
    StringSetting( RCString aKey, RCString defValue ) : Setting( aKey ),
      _defaultValue( defValue ),
      _value( defValue )
      {}
    virtual ~StringSetting() {}

    virtual bool isStringSetting () const { return true; }

    virtual OK_ERROR setValue( RCString newValue )
      { _value = newValue; return OK; }
    virtual OK_ERROR setValue( double newValue )
      { return setValue( String(newValue) ); }
    virtual OK_ERROR setValue( bool   newValue )
      { return setValue( String(newValue) ); }
    virtual OK_ERROR setValue( int    newValue )
      { return setValue( String(newValue) ); }

    String value        () const { return _value;        }
    String defaultValue () const { return _defaultValue; }

    virtual String stringValue    () const { return _value; }
    virtual void restoreDefault() { setValue(_defaultValue); }

    /*    Object'a */
    virtual classType   isA     () const { return stringSettingClass; }
    virtual const char* nameOf  () const { return "StringSetting"; }
    virtual bool        isEqual ( RCObject object ) const;

    /*  "" Clonable */
    virtual PClonable clone() const { return new StringSetting(ME); }

  private :

    const String _defaultValue;
    /* */ String _value;

  };

class BooleanSetting : public Setting
  {

  public :

    /* / */
    BooleanSetting( RCString aKey, bool defValue ) : Setting( aKey ),
      _defaultValue( defValue ),
      _value( defValue )
      {}
    virtual ~BooleanSetting() {}

    virtual bool isBooleanSetting () const { return true; }

    virtual OK_ERROR setValue( RCString newValue )
      { _value = bool(newValue); return OK; }
    virtual OK_ERROR setValue( bool     newValue )
      { _value = newValue; return OK; }
    virtual OK_ERROR setValue( double   newValue )
      { setValue( bool(newValue) ); return ERROR; }
    virtual OK_ERROR setValue( int      newValue )
      { return setValue( bool(newValue) ); }

    bool value        () const { return _value;        }
    bool defaultValue () const { return _defaultValue; }

    virtual String stringValue    () const { return String(_value); }
    virtual void restoreDefault() { setValue(_defaultValue); }

    /*    Object'a */
    virtual classType   isA     () const { return booleanSettingClass; }
    virtual const char* nameOf  () const { return "BooleanSetting"; }
    virtual bool        isEqual ( RCObject object ) const;

    /*  "" Clonable */
    virtual PClonable clone() const { return new BooleanSetting(ME); }

  private :

    const bool _defaultValue;
    /* */ bool _value;

  };

class IntegerSetting : public Setting
  {

  public :

    /* / */
    IntegerSetting( RCString aKey, int defValue ) : Setting( aKey ),
      _defaultValue( defValue ),
      _value( defValue )
      {}
    virtual ~IntegerSetting() {}

    virtual bool isIntegerSetting () const { return true; }

    virtual OK_ERROR setValue( RCString newValue )
      { _value = int(newValue); return OK; }
    virtual OK_ERROR setValue( int      newValue )
      { _value = newValue; return OK; }
    virtual OK_ERROR setValue( double   newValue )
      { setValue( int(newValue) ); return ERROR; }
    virtual OK_ERROR setValue( bool     newValue )
      { setValue( int(newValue) ); return ERROR; }

    int value        () const { return _value;        }
    int defaultValue () const { return _defaultValue; }

    virtual String stringValue    () const { return String(_value); }
    virtual void restoreDefault() { setValue(_defaultValue); }

    /*    Object'a */
    virtual classType   isA     () const { return integerSettingClass; }
    virtual const char* nameOf  () const { return "IntegerSetting"; }
    virtual bool        isEqual ( RCObject object ) const;

    /*  "" Clonable */
    virtual PClonable clone() const { return new IntegerSetting(ME); }

  private :

    const int _defaultValue;
    /* */ int _value;

  };

class Settings : public SrtArray
  {

  public :

    /* / */
    Settings()              : SrtArray()                   {}
    Settings( uint aDelta ) : SrtArray(aDelta)             {}
    Settings( RCSettings settings ) : SrtArray(settings) {}
    virtual ~Settings() {}

    /*   */
    OK_ERROR set( RCString key, double      value );
    OK_ERROR set( RCString key, RCString    value );
    OK_ERROR set( RCString key, bool        value );
    OK_ERROR set( RCString key, int         value );
    OK_ERROR set( RCString key, const char* value )
      { return set( key, String(value) ); }
    OK_ERROR set( RCString settingLine );
    OK_ERROR set( RCSetting s )
      { return contains(s.key()) ? set(s.key(),s.stringValue()) : (add(s),OK); }
    OK_ERROR set( RCSettings settings );

    double getRealSettingValue           ( RCString key ) const;
    String getStringSettingValue         ( RCString key ) const;
    bool   getBooleanSettingValue        ( RCString key ) const;
    int    getIntegerSettingValue        ( RCString key ) const;
    double getRealSettingDefaultValue    ( RCString key ) const;
    String getStringSettingDefaultValue  ( RCString key ) const;
    bool   getBooleanSettingDefaultValue ( RCString key ) const;
    int    getIntegerSettingDefaultValue ( RCString key ) const;

    void restoreDefaults();

    bool contains ( RCString  key ) const { return indexOf(key)<fBI; }
    int  indexOf  ( RCString  key ) const;

    bool contains ( RCSetting s   ) const { return SrtArray::contains(s);     }
    int  indexOf  ( RCSetting s   ) const { return SrtArray::firstIndexOf(s); }
    int  add      ( RCSetting s   )       { return SrtArray::add(s);          }
    int  remove   ( RCSetting s   )       { return SrtArray::remove(s);       }

    RSetting elementAt( int i )
      { return RSetting(SrtArray::elementAt(i)); }
    RCSetting elementAt( int i ) const
      { return RCSetting(SrtArray::elementAt(i)); }
    RCSetting firstElement() const
      { return RCSetting(SrtArray::firstElement()); }
    RCSetting lastElement() const
      { return RCSetting(SrtArray::lastElement()); }

    /*   */
    RCSetting operator [] ( int i ) const { return elementAt(i); }
    RSettings operator =  ( RCSettings settings )
      { return RSettings(SrtArray::operator=(settings)); }

    /*    Object'a */
    virtual classType   isA     () const { return settingsClass; }
    virtual const char* nameOf  () const { return "Settings"; }
    virtual void        printOn ( ostream& os ) const;

    /*  "" Clonable */
    virtual PClonable clone() const { return new Settings(ME); }

  protected :

    static PSetting parseSetting( RCString settingLine );

  };


inline RCRealSetting    Setting::toRealSetting    () const
  {
    if ( !this->isRealSetting() )
      cerr << "Error: illegal cast form " << nameOf() << " to RealSetting.\n";
    return RRealSetting(ME);
  }
inline RCStringSetting  Setting::toStringSetting  () const
  {
    if ( !this->isStringSetting() )
      cerr << "Error: illegal cast form " << nameOf() << " to StringSetting.\n";
    return RStringSetting(ME);
  }
inline RCBooleanSetting Setting::toBooleanSetting () const
  {
    if ( !this->isBooleanSetting() )
      cerr << "Error: illegal cast form " << nameOf() << " to BooleanSetting.\n";
    return RBooleanSetting(ME);
  }
inline RCIntegerSetting Setting::toIntegerSetting () const
  {
    if ( !this->isIntegerSetting() )
      cerr << "Error: illegal cast form " << nameOf() << " to IntegerSetting.\n";
    return RIntegerSetting(ME);
  }

inline double Setting::realValue    () const
  { return toRealSetting().value(); }
inline bool   Setting::booleanValue () const
  { return toBooleanSetting().value(); }
inline int    Setting::integerValue () const
  { return toIntegerSetting().value(); }

inline bool RealSetting::isEqual( RCObject object ) const
  { return object.isA() == realSettingClass &&
           RCSetting(object).key() == key() &&
           RCRealSetting(object).value() == value(); }
inline bool StringSetting::isEqual( RCObject object ) const
  { return object.isA() == stringSettingClass &&
           RCSetting(object).key() == key() &&
           RCStringSetting(object).value() == value(); }
inline bool BooleanSetting::isEqual ( RCObject object ) const
  { return object.isA() == booleanSettingClass &&
           RCSetting(object).key() == key() &&
           RCBooleanSetting(object).value() == value(); }
inline bool IntegerSetting::isEqual( RCObject object ) const
  { return object.isA() == integerSettingClass &&
           RCSetting(object).key() == key() &&
           RCIntegerSetting(object).value() == value(); }

inline OK_ERROR Settings::set( RCString key, double value )
  { int i=indexOf(key);
    return indexInRange(i) ? elementAt(i).setValue(value) : ERROR; }
inline OK_ERROR Settings::set( RCString key, RCString value )
  { int i=indexOf(key);
    return indexInRange(i) ? elementAt(i).setValue(value) : ERROR; }
inline OK_ERROR Settings::set( RCString key, bool value )
  { int i=indexOf(key);
    return indexInRange(i) ? elementAt(i).setValue(value) : ERROR; }
inline OK_ERROR Settings::set( RCString key, int value )
  { int i=indexOf(key);
    return indexInRange(i) ? elementAt(i).setValue(value) : ERROR; }

#endif  // __SETTINGS_H

