/* TInject.h -   TreeInjector  TreeInjection */

#ifndef __TINJECT_H
#define __TINJECT_H

#include "ClArray.h"
#include "GeneTree.h"  //   GeneTree
#include "SpecTree.h"  //   SpeciesTree
#include "GlSttngs.h"  //     

_CLASSDEF(TreeInjector)
_CLASSDEF(TreeInjection)

/*       */
class TreeInjector
  {

  public :

    TreeInjector( RCClArray sources, RCSpeciesTree target );
    ~TreeInjector() {}

    double inject();  //  !
    void   reset();   //  

    RCClArray      injections         () const { return injs; }
    RCClArray      sources            () const { return srcs; }
    RCSpeciesTree  target             () const { return trgt; }
    bool           injectionsComplete () const;

    OK_ERROR defineLengths();

  private :

    RCTreeInjection injection( int i ) const { return RCTreeInjection(injs[i]); }
    RTreeInjection  injection( int i )       { return RTreeInjection(injs[i]); }
    static void setNullLengths( RSpeciesTree sTree );

    const int     nOfSrcs;  //   
    ClArray       injs;     // 
    RCClArray     srcs;     // 
    RCSpeciesTree trgt;     // 
    bool          done;     //  

  };

/*       */
class TreeInjection : public GeneTree
  {

  public :

    TreeInjection( RCGeneTree      source, double cR );
    TreeInjection( RCTreeInjection aTree  );
    virtual ~TreeInjection() {}

    double injectTo( RCSpeciesTree target );  //   ...
    void   reset();                           //  

    RSpeciesTree image       () const { return *img; }
    bool   hasImage          () const { return img!=NULL; }
    double numberOfPassed    () const { return nOfPassed; }
    double numberOfConverged () const { return nOfCnvrgd; }
    double sumOfPassed       () const { return sumPassed; }
    double sumOfConverged    () const { return sumCnvrgd; }
    double quality           () const
      {  
//        IMPORT_REAL( costRatio, "CostRatio" );
        return sumPassed*costRatio+sumCnvrgd;
      }
    bool complete () const { return hasImage(); }

    RTreeInjection parent()
      { return RTreeInjection(GeneTree::parent()); }
    RCTreeInjection parent() const
      { return RTreeInjection(GeneTree::parent()); }
    RTreeInjection child( uint number )
      { return RTreeInjection(GeneTree::child(number)); }
    RCTreeInjection child( uint number ) const
      { return RTreeInjection(GeneTree::child(number)); }

    OK_ERROR defineLengths();

    /*    Object'a */
    virtual classType  isA      () const { return treeInjectionClass; }
    virtual const char* nameOf  () const { return "TreeInjection"; }
    virtual void        printOn ( ostream& os ) const;

    /*  "" Clonable */
    virtual Clonable* clone() const { return new TreeInjection(ME); }

    /*  "" Sortable */
    virtual bool isLessThan( RCSortable s ) const
      { return ( isA()==s.isA() && quality()<RCTreeInjection(s).quality() ); }

    virtual RTreeInjection operator = ( RCTreeInjection originalTree );
    virtual RGeneTree      operator = ( RCGeneTree      originalTree )
      { return RGeneTree( operator=( RCTree(originalTree) ) ); }
    virtual RTree          operator = ( RCTree          originalTree );

  private :

    PSpeciesTree img;        // 
    double       nOfPassed;  //  
    double       nOfCnvrgd;  //  
    double       sumPassed;  //  
    double       sumCnvrgd;  //  

    int injectLeaves( RCSpeciesTree target );  //  

    virtual int addChild( RCTree childToAdd );

    static RTreeInjection  treeToTI( RTree  tree );
    static RCTreeInjection treeToTI( RCTree tree );

    friend class TreeInjector;

  public :

    const double costRatio;

  };

inline TreeInjection::TreeInjection( RCGeneTree source, double cR ) :
  GeneTree( source.name(), source.distanceFromParent() ),
  img ( NULL ),
  nOfPassed(0), nOfCnvrgd(0),
  sumPassed(0), sumCnvrgd(0),
  costRatio(cR)
  { /*tlog<<'*';tlog.flush();*/ ME = source; }
inline TreeInjection::TreeInjection( RCTreeInjection source ) :
  GeneTree( source.name(), source.distanceFromParent() ),
  img ( NULL ),
  nOfPassed(0), nOfCnvrgd(0),
  sumPassed(0), sumCnvrgd(0),
  costRatio(source.costRatio)
  { ME = source; }

inline int TreeInjection::addChild( RCTree childToAdd )
  {
    if ( childToAdd.contains(ME) )
      return ERROR;

    return ( childToAdd.isA() == treeInjectionClass ?
             addChildPtr( new TreeInjection(RCTreeInjection(childToAdd)) ) :
             addChildPtr( new TreeInjection(childToAdd,costRatio) ) );
  }

inline RTreeInjection TreeInjection::treeToTI( RTree tree )
  {
//    if ( tree.isA() != treeInjectionClass )
//      cerr << "!   " << tree.nameOf()
//           << "&  TreeInjection& !" << endl;
    return RTreeInjection(tree);
  }
inline RCTreeInjection TreeInjection::treeToTI( RCTree tree )
  {
//    if ( tree.isA() != treeInjectionClass )
//      cerr << "!   " << tree.nameOf()
//           << "&  TreeInjection& !" << endl;
    return RTreeInjection(tree);
  }

#endif  // __TINJECT_H

